@@ -215,8 +215,7 @@ static int __mark_stripe_bucket(struct btree_trans *trans,
 		sectors = -sectors;
 
 	if (!deleting) {
-		if (bch2_trans_inconsistent_on(a->stripe ||
-					       a->stripe_redundancy, trans,
+		if (bch2_trans_inconsistent_on(a->stripe, trans,
 				"bucket %llu:%llu gen %u data type %s dirty_sectors %u: multiple stripes using same bucket (%u, %llu)\n%s",
 				bucket.inode, bucket.offset, a->gen,
 				bch2_data_type_str(a->data_type),
@@ -234,8 +233,7 @@ static int __mark_stripe_bucket(struct btree_trans *trans,
 				(bch2_bkey_val_to_text(&buf, c, s.s_c), buf.buf)))
 			return bch_err_throw(c, mark_stripe);
 	} else {
-		if (bch2_trans_inconsistent_on(a->stripe != s.k->p.offset ||
-					       a->stripe_redundancy != s.v->nr_redundant, trans,
+		if (bch2_trans_inconsistent_on(a->stripe != s.k->p.offset, trans,
 				"bucket %llu:%llu gen %u: not marked as stripe when deleting stripe (got %u)\n%s",
 				bucket.inode, bucket.offset, a->gen,
 				a->stripe,
@@ -267,11 +265,9 @@ static int __mark_stripe_bucket(struct btree_trans *trans,
 
 	if (!deleting) {
 		a->stripe		= s.k->p.offset;
-		a->stripe_redundancy	= s.v->nr_redundant;
 		alloc_data_type_set(a, data_type);
 	} else {
 		a->stripe		= 0;
-		a->stripe_redundancy	= 0;
 		alloc_data_type_set(a, BCH_DATA_user);
 	}
 
@@ -437,7 +433,7 @@ int bch2_trigger_stripe(struct btree_trans *trans,
 			struct disk_accounting_pos acc;
 			memset(&acc, 0, sizeof(acc));
 			acc.type = BCH_DISK_ACCOUNTING_replicas;
-			bch2_bkey_to_replicas(&acc.replicas, new);
+			bch2_bkey_to_replicas(c, &acc.replicas, new);
 			try(bch2_disk_accounting_mod(trans, &acc, &sectors, 1, gc));
 
 			if (gc)
@@ -451,18 +447,24 @@ int bch2_trigger_stripe(struct btree_trans *trans,
 			struct disk_accounting_pos acc;
 			memset(&acc, 0, sizeof(acc));
 			acc.type = BCH_DISK_ACCOUNTING_replicas;
-			bch2_bkey_to_replicas(&acc.replicas, old);
+			bch2_bkey_to_replicas(c, &acc.replicas, old);
 			try(bch2_disk_accounting_mod(trans, &acc, &sectors, 1, gc));
 		}
 
 		try(mark_stripe_buckets(trans, old, new, flags));
 	}
 
+	if ((flags & (BTREE_TRIGGER_atomic|BTREE_TRIGGER_gc)) == BTREE_TRIGGER_atomic) {
+		if (new_s && stripe_lru_pos(new_s) == 1)
+			bch2_do_stripe_deletes(c);
+	}
+
 	return 0;
 }
 
 /* returns blocknr in stripe that we matched: */
-static const struct bch_extent_ptr *bkey_matches_stripe(struct bch_stripe *s,
+static const struct bch_extent_ptr *bkey_matches_stripe(const struct bch_fs *c,
+							struct bch_stripe *s,
 						struct bkey_s_c k, unsigned *block)
 {
 	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
@@ -479,7 +481,7 @@ static const struct bch_extent_ptr *bkey_matches_stripe(struct bch_stripe *s,
 	return NULL;
 }
 
-static bool extent_has_stripe_ptr(struct bkey_s_c k, u64 idx)
+static bool extent_has_stripe_ptr(const struct bch_fs *c, struct bkey_s_c k, u64 idx)
 {
 	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
 	const union bch_extent_entry *entry;
@@ -508,6 +510,8 @@ static void ec_stripe_buf_exit(struct ec_stripe_buf *buf)
 	}
 }
 
+DEFINE_FREE(ec_stripe_buf_free, struct ec_stripe_buf *, ec_stripe_buf_exit(_T); kfree(_T));
+
 /* XXX: this is a non-mempoolified memory allocation: */
 static int ec_stripe_buf_init(struct bch_fs *c,
 			      struct ec_stripe_buf *buf,
@@ -531,14 +535,13 @@ static int ec_stripe_buf_init(struct bch_fs *c,
 
 	for (i = 0; i < v->nr_blocks; i++) {
 		buf->data[i] = kvmalloc(buf->size << 9, GFP_KERNEL);
-		if (!buf->data[i])
-			goto err;
+		if (!buf->data[i]) {
+			ec_stripe_buf_exit(buf);
+			return bch_err_throw(c, ENOMEM_stripe_buf);
+		}
 	}
 
 	return 0;
-err:
-	ec_stripe_buf_exit(buf);
-	return bch_err_throw(c, ENOMEM_stripe_buf);
 }
 
 /* Checksumming: */
@@ -783,80 +786,62 @@ static int get_stripe_key_trans(struct btree_trans *trans, u64 idx,
 	return 0;
 }
 
+static int stripe_reconstruct_err(struct bch_fs *c, struct bkey_s_c orig_k, const char *msg)
+{
+	CLASS(printbuf, msgbuf)();
+	bch2_bkey_val_to_text(&msgbuf, c, orig_k);
+	bch_err_ratelimited(c, "error doing reconstruct read: %s\n  %s", msg, msgbuf.buf);
+	return bch_err_throw(c, stripe_reconstruct);
+}
+
 /* recovery read path: */
 int bch2_ec_read_extent(struct btree_trans *trans, struct bch_read_bio *rbio,
 			struct bkey_s_c orig_k)
 {
 	struct bch_fs *c = trans->c;
-	struct ec_stripe_buf *buf = NULL;
-	struct closure cl;
-	struct bch_stripe *v;
-	unsigned i, offset;
-	const char *msg = NULL;
-	CLASS(printbuf, msgbuf)();
-	int ret = 0;
-
-	closure_init_stack(&cl);
 
 	BUG_ON(!rbio->pick.has_ec);
 
-	buf = kzalloc(sizeof(*buf), GFP_NOFS);
+	struct ec_stripe_buf *buf __free(ec_stripe_buf_free) = kzalloc(sizeof(*buf), GFP_NOFS);
 	if (!buf)
 		return bch_err_throw(c, ENOMEM_ec_read_extent);
 
-	ret = lockrestart_do(trans, get_stripe_key_trans(trans, rbio->pick.ec.idx, buf));
-	if (ret) {
-		msg = "stripe not found";
-		goto err;
-	}
+	int ret = lockrestart_do(trans, get_stripe_key_trans(trans, rbio->pick.ec.idx, buf));
+	if (ret)
+		return stripe_reconstruct_err(c, orig_k, "stripe not found");
 
-	v = &bkey_i_to_stripe(&buf->key)->v;
+	struct bch_stripe *v = &bkey_i_to_stripe(&buf->key)->v;
+	if (!bch2_ptr_matches_stripe(v, rbio->pick))
+		return stripe_reconstruct_err(c, orig_k, "pointer doesn't match stripe");
 
-	if (!bch2_ptr_matches_stripe(v, rbio->pick)) {
-		msg = "pointer doesn't match stripe";
-		goto err;
-	}
-
-	offset = rbio->bio.bi_iter.bi_sector - v->ptrs[rbio->pick.ec.block].offset;
-	if (offset + bio_sectors(&rbio->bio) > le16_to_cpu(v->sectors)) {
-		msg = "read is bigger than stripe";
-		goto err;
-	}
+	unsigned offset = rbio->bio.bi_iter.bi_sector - v->ptrs[rbio->pick.ec.block].offset;
+	if (offset + bio_sectors(&rbio->bio) > le16_to_cpu(v->sectors))
+		return stripe_reconstruct_err(c, orig_k, "read is bigger than stripe");
 
 	ret = ec_stripe_buf_init(c, buf, offset, bio_sectors(&rbio->bio));
-	if (ret) {
-		msg = "-ENOMEM";
-		goto err;
-	}
+	if (ret)
+		return stripe_reconstruct_err(c, orig_k, "-ENOMEM");
 
-	for (i = 0; i < v->nr_blocks; i++)
+	struct closure cl;
+	closure_init_stack(&cl);
+
+	for (unsigned i = 0; i < v->nr_blocks; i++)
 		ec_block_io(c, buf, REQ_OP_READ, i, &cl);
 
 	closure_sync(&cl);
 
-	if (ec_nr_failed(buf) > v->nr_redundant) {
-		msg = "unable to read enough blocks";
-		goto err;
-	}
+	if (ec_nr_failed(buf) > v->nr_redundant)
+		return stripe_reconstruct_err(c, orig_k, "unable to read enough blocks");
 
 	ec_validate_checksums(c, buf);
 
 	ret = ec_do_recov(c, buf);
 	if (ret)
-		goto err;
+		return stripe_reconstruct_err(c, orig_k, "unable to read enough blocks");
 
 	memcpy_to_bio(&rbio->bio, rbio->bio.bi_iter,
 		      buf->data[rbio->pick.ec.block] + ((offset - buf->offset) << 9));
-out:
-	ec_stripe_buf_exit(buf);
-	kfree(buf);
-	return ret;
-err:
-	bch2_bkey_val_to_text(&msgbuf, c, orig_k);
-	bch_err_ratelimited(c,
-			    "error doing reconstruct read: %s\n  %s", msg, msgbuf.buf);
-	ret = bch_err_throw(c, stripe_reconstruct);
-	goto out;
+	return 0;
 }
 
 /* stripe bucket accounting: */
@@ -881,31 +866,87 @@ static int ec_stripe_mem_alloc(struct btree_trans *trans,
  * Hash table of open stripes:
  * Stripes that are being created or modified are kept in a hash table, so that
  * stripe deletion can skip them.
+ *
+ * Additionally, we have a hash table for buckets that have stripes being
+ * created, to avoid racing with rebalance:
  */
 
-static bool __bch2_stripe_is_open(struct bch_fs *c, u64 idx)
+static bool __bch2_bucket_has_new_stripe(struct bch_fs *c, u64 dev_bucket)
+{
+	unsigned hash = hash_64(dev_bucket, ilog2(ARRAY_SIZE(c->ec_stripes_new_buckets)));
+	struct ec_stripe_new_bucket *s;
+
+	hlist_for_each_entry(s, &c->ec_stripes_new_buckets[hash], hash)
+		if (s->dev_bucket == dev_bucket)
+			return true;
+	return false;
+}
+
+bool bch2_bucket_has_new_stripe(struct bch_fs *c, u64 dev_bucket)
+{
+	guard(spinlock)(&c->ec_stripes_new_lock);
+	return __bch2_bucket_has_new_stripe(c, dev_bucket);
+}
+
+static void stripe_new_bucket_add(struct bch_fs *c, struct ec_stripe_new_bucket *s, u64 dev_bucket)
+{
+	s->dev_bucket = dev_bucket;
+
+	unsigned hash = hash_64(dev_bucket, ilog2(ARRAY_SIZE(c->ec_stripes_new_buckets)));
+	hlist_add_head(&s->hash, &c->ec_stripes_new_buckets[hash]);
+}
+
+static void stripe_new_buckets_add(struct bch_fs *c, struct ec_stripe_new *s)
+{
+	unsigned nr_blocks = s->nr_data + s->nr_parity;
+
+	guard(spinlock)(&c->ec_stripes_new_lock);
+	for (unsigned i = 0; i < nr_blocks; i++) {
+		if (!s->blocks[i])
+			continue;
+
+		struct open_bucket *ob = c->open_buckets + s->blocks[i];
+		struct bpos bucket = POS(ob->dev, ob->bucket);
+
+		stripe_new_bucket_add(c, &s->buckets[i], bucket_to_u64(bucket));
+	}
+}
+
+static void stripe_new_buckets_del(struct bch_fs *c, struct ec_stripe_new *s)
+{
+	guard(spinlock)(&c->ec_stripes_new_lock);
+
+	struct bch_stripe *v = &bkey_i_to_stripe(&s->new_stripe.key)->v;
+	for (unsigned i = 0; i < v->nr_blocks; i++)
+		hlist_del_init(&s->buckets[i].hash);
+}
+
+static struct ec_stripe_handle *bch2_open_stripe_find(struct bch_fs *c, u64 idx)
 {
 	unsigned hash = hash_64(idx, ilog2(ARRAY_SIZE(c->ec_stripes_new)));
-	struct ec_stripe_new *s;
+	struct ec_stripe_handle *s;
 
 	hlist_for_each_entry(s, &c->ec_stripes_new[hash], hash)
 		if (s->idx == idx)
-			return true;
-	return false;
+			return s;
+	return NULL;
 }
 
 static bool bch2_stripe_is_open(struct bch_fs *c, u64 idx)
 {
 	guard(spinlock)(&c->ec_stripes_new_lock);
-	return __bch2_stripe_is_open(c, idx);
+	return bch2_open_stripe_find(c, idx) != NULL;
 }
 
-static bool bch2_try_open_stripe(struct bch_fs *c,
-				 struct ec_stripe_new *s,
-				 u64 idx)
+static bool bch2_stripe_handle_tryget(struct bch_fs *c,
+				      struct ec_stripe_handle *s,
+				      u64 idx)
 {
+	BUG_ON(s->idx);
+	BUG_ON(!idx);
+
 	guard(spinlock)(&c->ec_stripes_new_lock);
-	bool ret = !__bch2_stripe_is_open(c, idx);
+	bool ret = !bch2_open_stripe_find(c, idx);
 	if (ret) {
 		unsigned hash = hash_64(idx, ilog2(ARRAY_SIZE(c->ec_stripes_new)));
 
@@ -915,11 +956,13 @@ static bool bch2_try_open_stripe(struct bch_fs *c,
 	return ret;
 }
 
-static void bch2_stripe_close(struct bch_fs *c, struct ec_stripe_new *s)
+static void bch2_stripe_handle_put(struct bch_fs *c, struct ec_stripe_handle *s)
 {
-	BUG_ON(!s->idx);
+	if (!s->idx)
+		return;
 
 	guard(spinlock)(&c->ec_stripes_new_lock);
+	BUG_ON(bch2_open_stripe_find(c, s->idx) != s);
 	hlist_del_init(&s->hash);
 
 	s->idx = 0;
@@ -1030,12 +1073,25 @@ static int ec_stripe_key_update(struct btree_trans *trans,
 	return bch2_trans_update(trans, &iter, &new->k_i, 0);
 }
 
+struct stripe_update_bucket_stats {
+	u32			nr_bp_to_deleted;
+	u32			nr_no_match;
+	u32			nr_cached;
+	u32			nr_done;
+
+	u32			sectors_bp_to_deleted;
+	u32			sectors_no_match;
+	u32			sectors_cached;
+	u32			sectors_done;
+};
+
 static int ec_stripe_update_extent(struct btree_trans *trans,
 				   struct bch_dev *ca,
 				   struct bpos bucket, u8 gen,
 				   struct ec_stripe_buf *s,
 				   struct bkey_s_c_backpointer bp,
-				   struct bkey_buf *last_flushed)
+				   struct stripe_update_bucket_stats *stats,
+				   struct wb_maybe_flush *last_flushed)
 {
 	struct bch_stripe *v = &bkey_i_to_stripe(&s->key)->v;
 	struct bch_fs *c = trans->c;
@@ -1063,20 +1119,33 @@ static int ec_stripe_update_extent(struct btree_trans *trans,
 		 * extent no longer exists - we could flush the btree
 		 * write buffer and retry to verify, but no need:
 		 */
+		stats->nr_bp_to_deleted++;
+		stats->sectors_bp_to_deleted += bp.v->bucket_len;
+		count_event(c, ec_stripe_update_extent_fail);
 		return 0;
 	}
 
-	if (extent_has_stripe_ptr(k, s->key.k.p.offset))
+	if (extent_has_stripe_ptr(c, k, s->key.k.p.offset))
 		return 0;
 
 	unsigned block;
-	const struct bch_extent_ptr *ptr_c = bkey_matches_stripe(v, k, &block);
+	const struct bch_extent_ptr *ptr_c = bkey_matches_stripe(c, v, k, &block);
 	/*
 	 * It doesn't generally make sense to erasure code cached ptrs:
 	 * XXX: should we be incrementing a counter?
 	 */
-	if (!ptr_c || ptr_c->cached)
+	if (!ptr_c) {
+		stats->nr_no_match++;
+		stats->sectors_no_match += bp.v->bucket_len;
+		count_event(c, ec_stripe_update_extent_fail);
 		return 0;
+	}
+	if (ptr_c->cached) {
+		stats->nr_cached++;
+		stats->sectors_cached += bp.v->bucket_len;
+		count_event(c, ec_stripe_update_extent_fail);
+		return 0;
+	}
 
 	unsigned dev = v->ptrs[block].dev;
 
@@ -1093,10 +1162,10 @@ static int ec_stripe_update_extent(struct btree_trans *trans,
 
 	bch2_bkey_drop_ptrs_noerror(bkey_i_to_s(n), p, entry, p.ptr.dev != dev);
 
-	struct bch_extent_ptr *ec_ptr = bch2_bkey_has_device(bkey_i_to_s(n), dev);
+	struct bch_extent_ptr *ec_ptr = bch2_bkey_has_device(c, bkey_i_to_s(n), dev);
 	BUG_ON(!ec_ptr);
 
-	__extent_entry_insert(n,
+	__extent_entry_insert(c, n,
 			(union bch_extent_entry *) ec_ptr,
 			(union bch_extent_entry *) &stripe_ptr);
 
@@ -1106,6 +1175,14 @@ static int ec_stripe_update_extent(struct btree_trans *trans,
 	try(bch2_bkey_set_needs_rebalance(trans->c, &opts, n,
 					  SET_NEEDS_REBALANCE_other, 0));
 	try(bch2_trans_update(trans, &iter, n, 0));
+	try(bch2_trans_commit(trans, NULL, NULL,
+			BCH_TRANS_COMMIT_no_check_rw|
+			BCH_TRANS_COMMIT_no_enospc));
+
+	stats->nr_done++;
+	stats->sectors_done += bp.v->bucket_len;
+
+	count_event(c, ec_stripe_update_extent);
 
 	return 0;
 }
@@ -1123,15 +1200,14 @@ static int ec_stripe_update_bucket(struct btree_trans *trans, struct ec_stripe_b
 
 	struct bpos bucket_pos = PTR_BUCKET_POS(ca, &ptr);
 
-	struct bkey_buf last_flushed __cleanup(bch2_bkey_buf_exit);
-	bch2_bkey_buf_init(&last_flushed);
+	struct wb_maybe_flush last_flushed __cleanup(wb_maybe_flush_exit);
+	wb_maybe_flush_init(&last_flushed);
 
-	return for_each_btree_key_max_commit(trans, bp_iter, BTREE_ID_backpointers,
+	struct stripe_update_bucket_stats stats = {};
+
+	try(for_each_btree_key_max(trans, bp_iter, BTREE_ID_backpointers,
 			bucket_pos_to_bp_start(ca, bucket_pos),
-			bucket_pos_to_bp_end(ca, bucket_pos), 0, bp_k,
-			NULL, NULL,
-			BCH_TRANS_COMMIT_no_check_rw|
-			BCH_TRANS_COMMIT_no_enospc, ({
+			bucket_pos_to_bp_end(ca, bucket_pos), 0, bp_k, ({
 		if (bkey_ge(bp_k.k->p, bucket_pos_to_bp(ca, bpos_nosnap_successor(bucket_pos), 0)))
 			break;
 
@@ -1142,8 +1218,27 @@ static int ec_stripe_update_bucket(struct btree_trans *trans, struct ec_stripe_b
 		if (bp.v->btree_id == BTREE_ID_stripes)
 			continue;
 
-		ec_stripe_update_extent(trans, ca, bucket_pos, ptr.gen, s, bp, &last_flushed);
-	}));
+		wb_maybe_flush_inc(&last_flushed);
+		ec_stripe_update_extent(trans, ca, bucket_pos, ptr.gen, s, bp,
+					&stats, &last_flushed);
+	})));
+
+	if (trace_stripe_update_bucket_enabled()) {
+		CLASS(printbuf, buf)();
+
+		prt_printf(&buf, "bp_to_deleted:\t%u %u\n",
+			   stats.nr_bp_to_deleted, stats.sectors_bp_to_deleted);
+		prt_printf(&buf, "no_match:\t%u %u\n",
+			   stats.nr_no_match, stats.sectors_no_match);
+		prt_printf(&buf, "cached:\t%u %u\n",
+			   stats.nr_cached, stats.sectors_cached);
+		prt_printf(&buf, "done:\t%u %u\n",
+			   stats.nr_done, stats.sectors_done);
+
+		trace_stripe_update_bucket(c, buf.buf);
+	}
+
+	return 0;
 }
 
 static int ec_stripe_update_extents(struct bch_fs *c, struct ec_stripe_buf *s)
@@ -1190,102 +1285,100 @@ static void zero_out_rest_of_ec_bucket(struct bch_fs *c,
 
 void bch2_ec_stripe_new_free(struct bch_fs *c, struct ec_stripe_new *s)
 {
-	if (s->idx)
-		bch2_stripe_close(c, s);
+	stripe_new_buckets_del(c, s);
+	bch2_stripe_handle_put(c, &s->new_stripe_handle);
+	bch2_stripe_handle_put(c, &s->old_stripe_handle);
 	kfree(s);
 }
 
+static int __ec_stripe_create(struct ec_stripe_new *s)
+{
+	struct bch_fs *c = s->c;
+	struct bch_stripe *v = &bkey_i_to_stripe(&s->new_stripe.key)->v;
+	unsigned nr_data = v->nr_blocks - v->nr_redundant;
+
+	if (s->err) {
+		if (!bch2_err_matches(s->err, EROFS))
+			bch_err(c, "error creating stripe: error writing data buckets");
+		return s->err;
+	}
+
+	for (unsigned i = 0; i < nr_data; i++)
+		if (s->blocks[i]) {
+			struct open_bucket *ob = c->open_buckets + s->blocks[i];
+
+			if (ob->sectors_free)
+				zero_out_rest_of_ec_bucket(c, s, i, ob);
+		}
+
+	if (s->have_old_stripe) {
+		ec_validate_checksums(c, &s->old_stripe);
+
+		if (ec_do_recov(c, &s->old_stripe)) {
+			bch_err(c, "error creating stripe: error reading old stripe");
+			return bch_err_throw(c, ec_block_read);
+		}
+
+		for (unsigned i = 0; i < nr_data; i++)
+			if (stripe_blockcount_get(&bkey_i_to_stripe(&s->old_stripe.key)->v, i))
+				swap(s->new_stripe.data[i],
+				     s->old_stripe.data[i]);
+
+		ec_stripe_buf_exit(&s->old_stripe);
+	}
+
+	BUG_ON(!s->allocated);
+
+	ec_generate_ec(&s->new_stripe);
+	ec_generate_checksums(&s->new_stripe);
+
+	/* write p/q: */
+	for (unsigned i = nr_data; i < v->nr_blocks; i++)
+		ec_block_io(c, &s->new_stripe, REQ_OP_WRITE, i, &s->iodone);
+	closure_sync(&s->iodone);
+
+	if (ec_nr_failed(&s->new_stripe)) {
+		bch_err(c, "error creating stripe: error writing redundancy buckets");
+		return bch_err_throw(c, ec_block_write);
+	}
+
+	try(bch2_trans_commit_do(c, &s->res, NULL,
+		BCH_TRANS_COMMIT_no_check_rw|
+		BCH_TRANS_COMMIT_no_enospc,
+		ec_stripe_key_update(trans,
+				     s->have_old_stripe
+				     ? bkey_i_to_stripe(&s->old_stripe.key)
+				     : NULL,
+				     bkey_i_to_stripe(&s->new_stripe.key))));
+	try(ec_stripe_update_extents(c, &s->new_stripe));
+
+	return 0;
+}
+
 /*
  * data buckets of new stripe all written: create the stripe
  */
 static void ec_stripe_create(struct ec_stripe_new *s)
 {
 	struct bch_fs *c = s->c;
-	struct open_bucket *ob;
 	struct bch_stripe *v = &bkey_i_to_stripe(&s->new_stripe.key)->v;
-	unsigned i, nr_data = v->nr_blocks - v->nr_redundant;
-	int ret;
+	unsigned nr_data = v->nr_blocks - v->nr_redundant;
 
 	BUG_ON(s->h->s == s);
 
 	closure_sync(&s->iodone);
 
-	if (!s->err) {
-		for (i = 0; i < nr_data; i++)
-			if (s->blocks[i]) {
-				ob = c->open_buckets + s->blocks[i];
+	int ret = __ec_stripe_create(s);
+	if (ret && !s->err)
+		s->err = ret;
 
-				if (ob->sectors_free)
-					zero_out_rest_of_ec_bucket(c, s, i, ob);
-			}
-	}
-
-	if (s->err) {
-		if (!bch2_err_matches(s->err, EROFS))
-			bch_err(c, "error creating stripe: error writing data buckets");
-		ret = s->err;
-		goto err;
-	}
-
-	if (s->have_existing_stripe) {
-		ec_validate_checksums(c, &s->existing_stripe);
-
-		if (ec_do_recov(c, &s->existing_stripe)) {
-			bch_err(c, "error creating stripe: error reading existing stripe");
-			ret = bch_err_throw(c, ec_block_read);
-			goto err;
-		}
-
-		for (i = 0; i < nr_data; i++)
-			if (stripe_blockcount_get(&bkey_i_to_stripe(&s->existing_stripe.key)->v, i))
-				swap(s->new_stripe.data[i],
-				     s->existing_stripe.data[i]);
-
-		ec_stripe_buf_exit(&s->existing_stripe);
-	}
-
-	BUG_ON(!s->allocated);
-	BUG_ON(!s->idx);
-
-	ec_generate_ec(&s->new_stripe);
-
-	ec_generate_checksums(&s->new_stripe);
-
-	/* write p/q: */
-	for (i = nr_data; i < v->nr_blocks; i++)
-		ec_block_io(c, &s->new_stripe, REQ_OP_WRITE, i, &s->iodone);
-	closure_sync(&s->iodone);
-
-	if (ec_nr_failed(&s->new_stripe)) {
-		bch_err(c, "error creating stripe: error writing redundancy buckets");
-		ret = bch_err_throw(c, ec_block_write);
-		goto err;
-	}
-
-	ret = bch2_trans_commit_do(c, &s->res, NULL,
-		BCH_TRANS_COMMIT_no_check_rw|
-		BCH_TRANS_COMMIT_no_enospc,
-		ec_stripe_key_update(trans,
-				     s->have_existing_stripe
-				     ? bkey_i_to_stripe(&s->existing_stripe.key)
-				     : NULL,
-				     bkey_i_to_stripe(&s->new_stripe.key)));
-	bch_err_msg(c, ret, "creating stripe key");
-	if (ret)
-		goto err;
-
-	ret = ec_stripe_update_extents(c, &s->new_stripe);
-	bch_err_msg(c, ret, "error updating extents");
-	if (ret)
-		goto err;
-err:
-	trace_stripe_create(c, s->idx, ret);
+	trace_stripe_create(c, s->new_stripe.key.k.p.offset, ret);
 
 	bch2_disk_reservation_put(c, &s->res);
 
-	for (i = 0; i < v->nr_blocks; i++)
+	for (unsigned i = 0; i < v->nr_blocks; i++)
 		if (s->blocks[i]) {
-			ob = c->open_buckets + s->blocks[i];
+			struct open_bucket *ob = c->open_buckets + s->blocks[i];
 
 			if (i < nr_data) {
 				ob->ec = NULL;
@@ -1299,7 +1392,7 @@ err:
 		list_del(&s->list);
 	wake_up(&c->ec_stripe_new_wait);
 
-	ec_stripe_buf_exit(&s->existing_stripe);
+	ec_stripe_buf_exit(&s->old_stripe);
 	ec_stripe_buf_exit(&s->new_stripe);
 	closure_debug_destroy(&s->iodone);
 
@@ -1577,12 +1670,11 @@ __bch2_ec_stripe_head_get(struct btree_trans *trans,
 {
 	struct bch_fs *c = trans->c;
 	struct ec_stripe_head *h;
-	int ret;
 
 	if (!redundancy)
 		return NULL;
 
-	ret = bch2_trans_mutex_lock(trans, &c->ec_stripe_head_lock);
+	int ret = bch2_trans_mutex_lock(trans, &c->ec_stripe_head_lock);
 	if (ret)
 		return ERR_PTR(ret);
 
@@ -1622,7 +1714,7 @@ err:
 	return h;
 }
 
-static int new_stripe_alloc_buckets(struct btree_trans *trans,
+static int __new_stripe_alloc_buckets(struct btree_trans *trans,
 				    struct alloc_request *req,
 				    struct ec_stripe_head *h, struct ec_stripe_new *s,
 				    struct closure *cl)
@@ -1631,17 +1723,6 @@ static int new_stripe_alloc_buckets(struct btree_trans *trans,
 	struct open_bucket *ob;
 	struct bch_stripe *v = &bkey_i_to_stripe(&s->new_stripe.key)->v;
 	unsigned i, j, nr_have_parity = 0, nr_have_data = 0;
-	int ret = 0;
-
-	req->scratch_data_type		= req->data_type;
-	req->scratch_ptrs		= req->ptrs;
-	req->scratch_nr_replicas	= req->nr_replicas;
-	req->scratch_nr_effective	= req->nr_effective;
-	req->scratch_have_cache		= req->have_cache;
-	req->scratch_devs_may_alloc	= req->devs_may_alloc;
-
-	req->devs_may_alloc	= h->devs;
-	req->have_cache		= true;
 
 	BUG_ON(v->nr_blocks	!= s->nr_data + s->nr_parity);
 	BUG_ON(v->nr_redundant	!= s->nr_parity);
@@ -1675,7 +1756,7 @@ static int new_stripe_alloc_buckets(struct btree_trans *trans,
 		req->nr_effective	= nr_have_parity;
 		req->data_type		= BCH_DATA_parity;
 
-		ret = bch2_bucket_alloc_set_trans(trans, req, &h->parity_stripe, cl);
+		int ret = bch2_bucket_alloc_set_trans(trans, req, &h->parity_stripe, cl);
 
 		open_bucket_for_each(c, &req->ptrs, ob, i) {
 			j = find_next_zero_bit(s->blocks_gotten,
@@ -1689,7 +1770,7 @@ static int new_stripe_alloc_buckets(struct btree_trans *trans,
 		}
 
 		if (ret)
-			goto err;
+			return ret;
 	}
 
 	req->ptrs.nr = 0;
@@ -1698,7 +1779,7 @@ static int new_stripe_alloc_buckets(struct btree_trans *trans,
 		req->nr_effective	= nr_have_data;
 		req->data_type		= BCH_DATA_user;
 
-		ret = bch2_bucket_alloc_set_trans(trans, req, &h->block_stripe, cl);
+		int ret = bch2_bucket_alloc_set_trans(trans, req, &h->block_stripe, cl);
 
 		open_bucket_for_each(c, &req->ptrs, ob, i) {
 			j = find_next_zero_bit(s->blocks_gotten,
@@ -1711,9 +1792,34 @@ static int new_stripe_alloc_buckets(struct btree_trans *trans,
 		}
 
 		if (ret)
-			goto err;
+			return ret;
 	}
-err:
+
+	return 0;
+}
+
+static int new_stripe_alloc_buckets(struct btree_trans *trans,
+				    struct alloc_request *req,
+				    struct ec_stripe_head *h, struct ec_stripe_new *s,
+				    struct closure *cl)
+{
+	struct bch_stripe *v = &bkey_i_to_stripe(&s->new_stripe.key)->v;
+
+	if (bitmap_weight(s->blocks_gotten, v->nr_blocks) == v->nr_blocks)
+		return 0;
+
+	req->scratch_data_type		= req->data_type;
+	req->scratch_ptrs		= req->ptrs;
+	req->scratch_nr_replicas	= req->nr_replicas;
+	req->scratch_nr_effective	= req->nr_effective;
+	req->scratch_have_cache		= req->have_cache;
+	req->scratch_devs_may_alloc	= req->devs_may_alloc;
+
+	req->devs_may_alloc	= h->devs;
+	req->have_cache		= true;
+
+	int ret = __new_stripe_alloc_buckets(trans, req, h, s, cl);
+
 	req->data_type		= req->scratch_data_type;
 	req->ptrs		= req->scratch_ptrs;
 	req->nr_replicas	= req->scratch_nr_replicas;
@@ -1723,7 +1829,7 @@ err:
 	return ret;
 }
 
-static int __get_existing_stripe(struct btree_trans *trans,
+static int __get_old_stripe(struct btree_trans *trans,
 				 struct ec_stripe_head *head,
 				 struct ec_stripe_buf *stripe,
 				 u64 idx)
@@ -1745,7 +1851,7 @@ static int __get_existing_stripe(struct btree_trans *trans,
 	    s.v->algorithm		== head->algo &&
 	    s.v->nr_redundant		== head->redundancy &&
 	    le16_to_cpu(s.v->sectors)	== head->blocksize &&
-	    bch2_try_open_stripe(c, head->s, idx)) {
+	    bch2_stripe_handle_tryget(c, &head->s->old_stripe_handle, idx)) {
 		bkey_reassemble(&stripe->key, k);
 		return 1;
 	}
@@ -1753,23 +1859,23 @@ static int __get_existing_stripe(struct btree_trans *trans,
 	return 0;
 }
 
-static int init_new_stripe_from_existing(struct bch_fs *c, struct ec_stripe_new *s)
+static int init_new_stripe_from_old(struct bch_fs *c, struct ec_stripe_new *s)
 {
 	struct bch_stripe *new_v = &bkey_i_to_stripe(&s->new_stripe.key)->v;
-	struct bch_stripe *existing_v = &bkey_i_to_stripe(&s->existing_stripe.key)->v;
+	struct bch_stripe *old_v = &bkey_i_to_stripe(&s->old_stripe.key)->v;
 	unsigned i;
 
-	BUG_ON(existing_v->nr_redundant != s->nr_parity);
-	s->nr_data = existing_v->nr_blocks -
-		existing_v->nr_redundant;
+	BUG_ON(old_v->nr_redundant != s->nr_parity);
+	s->nr_data = old_v->nr_blocks -
+		old_v->nr_redundant;
 
-	int ret = ec_stripe_buf_init(c, &s->existing_stripe, 0, le16_to_cpu(existing_v->sectors));
+	int ret = ec_stripe_buf_init(c, &s->old_stripe, 0, le16_to_cpu(old_v->sectors));
 	if (ret) {
-		bch2_stripe_close(c, s);
+		bch2_stripe_handle_put(c, &s->old_stripe_handle);
 		return ret;
 	}
 
-	BUG_ON(s->existing_stripe.size != le16_to_cpu(existing_v->sectors));
+	BUG_ON(s->old_stripe.size != le16_to_cpu(old_v->sectors));
 
 	/*
 	 * Free buckets we initially allocated - they might conflict with
@@ -1782,22 +1888,22 @@ static int init_new_stripe_from_existing(struct bch_fs *c, struct ec_stripe_new
 	memset(s->blocks_gotten, 0, sizeof(s->blocks_gotten));
 	memset(s->blocks_allocated, 0, sizeof(s->blocks_allocated));
 
-	for (unsigned i = 0; i < existing_v->nr_blocks; i++) {
-		if (stripe_blockcount_get(existing_v, i)) {
+	for (unsigned i = 0; i < old_v->nr_blocks; i++) {
+		if (stripe_blockcount_get(old_v, i)) {
 			__set_bit(i, s->blocks_gotten);
 			__set_bit(i, s->blocks_allocated);
 		}
 
-		ec_block_io(c, &s->existing_stripe, READ, i, &s->iodone);
+		ec_block_io(c, &s->old_stripe, READ, i, &s->iodone);
 	}
 
-	bkey_copy(&s->new_stripe.key, &s->existing_stripe.key);
-	s->have_existing_stripe = true;
+	bkey_copy(&s->new_stripe.key, &s->old_stripe.key);
+	s->have_old_stripe = true;
 
 	return 0;
 }
 
-static int __bch2_ec_stripe_head_reuse(struct btree_trans *trans, struct ec_stripe_head *h,
+static int __bch2_ec_stripe_reuse(struct btree_trans *trans, struct ec_stripe_head *h,
 				       struct ec_stripe_new *s)
 {
 	struct bch_fs *c = trans->c;
@@ -1816,7 +1922,7 @@ static int __bch2_ec_stripe_head_reuse(struct btree_trans *trans, struct ec_stri
 			lru_pos(BCH_LRU_STRIPE_FRAGMENTATION, 2, 0),
 			lru_pos(BCH_LRU_STRIPE_FRAGMENTATION, 2, LRU_TIME_MAX),
 			0, lru_k, ret) {
-		ret = __get_existing_stripe(trans, h, &s->existing_stripe, lru_k.k->p.offset);
+		ret = __get_old_stripe(trans, h, &s->old_stripe, lru_k.k->p.offset);
 		if (ret)
 			break;
 	}
@@ -1827,28 +1933,21 @@ static int __bch2_ec_stripe_head_reuse(struct btree_trans *trans, struct ec_stri
 	if (ret)
 		return ret;
 
-	return init_new_stripe_from_existing(c, s);
+	return init_new_stripe_from_old(c, s);
 }
 
-static int __bch2_ec_stripe_head_reserve(struct btree_trans *trans, struct ec_stripe_head *h,
-					 struct ec_stripe_new *s)
+static int stripe_idx_alloc(struct btree_trans *trans, struct ec_stripe_new *s)
 {
+	/*
+	 * Allocate stripe slot
+	 * XXX: we're going to need a bitrange btree of free stripes
+	 */
 	struct bch_fs *c = trans->c;
 	struct bkey_s_c k;
 	struct bpos min_pos = POS(0, 1);
 	struct bpos start_pos = bpos_max(min_pos, POS(0, c->ec_stripe_hint));
 	int ret;
 
-	if (!s->res.sectors)
-		try(bch2_disk_reservation_get(c, &s->res,
-					h->blocksize,
-					s->nr_parity,
-					BCH_DISK_RESERVATION_NOFAIL));
-
-	/*
-	 * Allocate stripe slot
-	 * XXX: we're going to need a bitrange btree of free stripes
-	 */
 	for_each_btree_key_norestart(trans, iter, BTREE_ID_stripes, start_pos,
 			   BTREE_ITER_slots|BTREE_ITER_intent, k, ret) {
 		c->ec_stripe_hint = iter.pos.offset;
@@ -1865,20 +1964,82 @@ static int __bch2_ec_stripe_head_reserve(struct btree_trans *trans, struct ec_st
 		}
 
 		if (bkey_deleted(k.k) &&
-		    bch2_try_open_stripe(c, s, k.k->p.offset)) {
+		    bch2_stripe_handle_tryget(c, &s->new_stripe_handle, k.k->p.offset)) {
 			ret = ec_stripe_mem_alloc(trans, &iter);
 			if (ret)
-				bch2_stripe_close(c, s);
+				bch2_stripe_handle_put(c, &s->new_stripe_handle);
 			s->new_stripe.key.k.p = iter.pos;
 			break;
 		}
 	}
 
-	if (ret)
-		bch2_disk_reservation_put(c, &s->res);
 	return ret;
 }
 
+static int stripe_alloc_or_reuse(struct btree_trans *trans,
+				 struct alloc_request *req,
+				 struct closure *cl,
+				 struct ec_stripe_head *h,
+				 struct ec_stripe_new *s,
+				 bool *waiting)
+{
+	struct bch_fs *c = trans->c;
+
+	if (!s->have_old_stripe) {
+		/* First, try to allocate a full stripe: */
+		enum bch_watermark saved_watermark = BCH_WATERMARK_stripe;
+		swap(req->watermark, saved_watermark);
+		int ret = new_stripe_alloc_buckets(trans, req, h, s, NULL) ?:
+			stripe_idx_alloc(trans, s);
+		swap(req->watermark, saved_watermark);
+
+		if (ret) {
+			if (bch2_err_matches(ret, BCH_ERR_transaction_restart) ||
+			    bch2_err_matches(ret, ENOMEM))
+				return ret;
+
+			/*
+			 * Not enough buckets available for a full stripe: we must reuse an
+			 * oldstripe:
+			 */
+			while (1) {
+				ret = __bch2_ec_stripe_reuse(trans, h, s);
+				if (!ret)
+					break;
+				if (*waiting || !cl || ret != -BCH_ERR_stripe_alloc_blocked)
+					return ret;
+
+				if (req->watermark == BCH_WATERMARK_copygc) {
+					try(new_stripe_alloc_buckets(trans, req, h, s, NULL));
+					try(stripe_idx_alloc(trans, s));
+					break;
+				}
+
+				/* XXX freelist_wait? */
+				closure_wait(&c->freelist_wait, cl);
+				*waiting = true;
+			}
+		}
+	}
+
+	/*
+	 * Retry allocating buckets, with the watermark for this
+	 * particular write:
+	 */
+	try(new_stripe_alloc_buckets(trans, req, h, s, cl));
+	try(ec_stripe_buf_init(c, &s->new_stripe, 0, h->blocksize));
+
+	if (!s->have_old_stripe && !s->res.sectors)
+		bch2_disk_reservation_get(c, &s->res,
+					  h->blocksize,
+					  s->nr_parity,
+					  BCH_DISK_RESERVATION_NOFAIL);
+
+	stripe_new_buckets_add(c, s);
+	s->allocated = true;
+	return 0;
+}
+
 struct ec_stripe_head *bch2_ec_stripe_head_get(struct btree_trans *trans,
 					       struct alloc_request *req,
 					       unsigned algo,
@@ -1888,7 +2049,6 @@ struct ec_stripe_head *bch2_ec_stripe_head_get(struct btree_trans *trans,
 	unsigned redundancy = req->nr_replicas - 1;
 	unsigned disk_label = 0;
 	struct target t = target_decode(req->target);
-	bool waiting = false;
 	int ret;
 
 	if (t.type == TARGET_GROUP) {
@@ -1917,76 +2077,20 @@ struct ec_stripe_head *bch2_ec_stripe_head_get(struct btree_trans *trans,
 	}
 
 	struct ec_stripe_new *s = h->s;
+	if (!s->allocated) {
+		bool waiting = false;
+		ret = stripe_alloc_or_reuse(trans, req, cl, h, s, &waiting);
+		if (waiting &&
+		    !bch2_err_matches(ret, BCH_ERR_operation_blocked))
+			closure_wake_up(&c->freelist_wait);
 
-	if (s->allocated)
-		goto allocated;
-
-	if (s->have_existing_stripe)
-		goto alloc_existing;
-
-	/* First, try to allocate a full stripe: */
-	enum bch_watermark saved_watermark = BCH_WATERMARK_stripe;
-	swap(req->watermark, saved_watermark);
-	ret =   new_stripe_alloc_buckets(trans, req, h, s, NULL) ?:
-		__bch2_ec_stripe_head_reserve(trans, h, s);
-	swap(req->watermark, saved_watermark);
-
-	if (!ret)
-		goto allocate_buf;
-	if (bch2_err_matches(ret, BCH_ERR_transaction_restart) ||
-	    bch2_err_matches(ret, ENOMEM))
-		goto err;
-
-	/*
-	 * Not enough buckets available for a full stripe: we must reuse an
-	 * existing stripe:
-	 */
-	while (1) {
-		ret = __bch2_ec_stripe_head_reuse(trans, h, s);
-		if (!ret)
-			break;
-		if (waiting || !cl || ret != -BCH_ERR_stripe_alloc_blocked)
+		if (ret)
 			goto err;
-
-		if (req->watermark == BCH_WATERMARK_copygc) {
-			ret =   new_stripe_alloc_buckets(trans, req, h, s, NULL) ?:
-				__bch2_ec_stripe_head_reserve(trans, h, s);
-			if (ret)
-				goto err;
-			goto allocate_buf;
-		}
-
-		/* XXX freelist_wait? */
-		closure_wait(&c->freelist_wait, cl);
-		waiting = true;
 	}
-
-	if (waiting)
-		closure_wake_up(&c->freelist_wait);
-alloc_existing:
-	/*
-	 * Retry allocating buckets, with the watermark for this
-	 * particular write:
-	 */
-	ret = new_stripe_alloc_buckets(trans, req, h, s, cl);
-	if (ret)
-		goto err;
-
-allocate_buf:
-	ret = ec_stripe_buf_init(c, &s->new_stripe, 0, h->blocksize);
-	if (ret)
-		goto err;
-
-	s->allocated = true;
-allocated:
-	BUG_ON(!s->idx);
 	BUG_ON(!s->new_stripe.data[0]);
 	BUG_ON(trans->restarted);
 	return h;
 err:
-	if (waiting &&
-	    !bch2_err_matches(ret, BCH_ERR_operation_blocked))
-		closure_wake_up(&c->freelist_wait);
 	bch2_ec_stripe_head_put(c, h);
 	return ERR_PTR(ret);
 }
@@ -2014,7 +2118,7 @@ int bch2_invalidate_stripe_to_dev(struct btree_trans *trans,
 
 	memset(&acc, 0, sizeof(acc));
 	acc.type = BCH_DISK_ACCOUNTING_replicas;
-	bch2_bkey_to_replicas(&acc.replicas, bkey_i_to_s_c(&s->k_i));
+	bch2_bkey_to_replicas(c, &acc.replicas, bkey_i_to_s_c(&s->k_i));
 	acc.replicas.data_type = BCH_DATA_user;
 	try(bch2_disk_accounting_mod(trans, &acc, &sectors, 1, false));
 
@@ -2053,7 +2157,7 @@ int bch2_invalidate_stripe_to_dev(struct btree_trans *trans,
 
 	memset(&acc, 0, sizeof(acc));
 	acc.type = BCH_DISK_ACCOUNTING_replicas;
-	bch2_bkey_to_replicas(&acc.replicas, bkey_i_to_s_c(&s->k_i));
+	bch2_bkey_to_replicas(c, &acc.replicas, bkey_i_to_s_c(&s->k_i));
 	acc.replicas.data_type = BCH_DATA_user;
 	return bch2_disk_accounting_mod(trans, &acc, &sectors, 1, false);
 }
@@ -2095,6 +2199,23 @@ int bch2_dev_remove_stripes(struct bch_fs *c, unsigned dev_idx,
 
 /* startup/shutdown */
 
+static bool should_cancel_stripe(struct bch_fs *c, struct ec_stripe_new *s, struct bch_dev *ca)
+{
+	if (!ca)
+		return true;
+
+	for (unsigned i = 0; i < bkey_i_to_stripe(&s->new_stripe.key)->v.nr_blocks; i++) {
+		if (!s->blocks[i])
+			continue;
+
+		struct open_bucket *ob = c->open_buckets + s->blocks[i];
+		if (ob->dev == ca->dev_idx)
+			return true;
+	}
+
+	return false;
+}
+
 static void __bch2_ec_stop(struct bch_fs *c, struct bch_dev *ca)
 {
 	struct ec_stripe_head *h;
@@ -2102,23 +2223,8 @@ static void __bch2_ec_stop(struct bch_fs *c, struct bch_dev *ca)
 	guard(mutex)(&c->ec_stripe_head_lock);
 	list_for_each_entry(h, &c->ec_stripe_head_list, list) {
 		guard(mutex)(&h->lock);
-		if (!h->s)
-			continue;
-
-		if (!ca)
-			goto found;
-
-		for (unsigned i = 0; i < bkey_i_to_stripe(&h->s->new_stripe.key)->v.nr_blocks; i++) {
-			if (!h->s->blocks[i])
-				continue;
-
-			struct open_bucket *ob = c->open_buckets + h->s->blocks[i];
-			if (ob->dev == ca->dev_idx)
-				goto found;
-		}
-		continue;
-found:
-		ec_stripe_new_cancel(c, h, -BCH_ERR_erofs_no_writes);
+		if (h->s && should_cancel_stripe(c, h->s, ca))
+			ec_stripe_new_cancel(c, h, -BCH_ERR_erofs_no_writes);
 	}
 }
 
@@ -2154,7 +2260,7 @@ static void bch2_new_stripe_to_text(struct printbuf *out, struct bch_fs *c,
 				    struct ec_stripe_new *s)
 {
 	prt_printf(out, "\tidx %llu blocks %u+%u allocated %u ref %u %u %s obs",
-		   s->idx, s->nr_data, s->nr_parity,
+		   s->new_stripe.key.k.p.offset, s->nr_data, s->nr_parity,
 		   bitmap_weight(s->blocks_allocated, s->nr_data),
 		   atomic_read(&s->ref[STRIPE_REF_io]),
 		   atomic_read(&s->ref[STRIPE_REF_stripe]),
@@ -2243,7 +2349,7 @@ int bch2_fs_ec_init(struct bch_fs *c)
 
 static int bch2_check_stripe_to_lru_ref(struct btree_trans *trans,
 					struct bkey_s_c k,
-					struct bkey_buf *last_flushed)
+					struct wb_maybe_flush *last_flushed)
 {
 	if (k.k->type != KEY_TYPE_stripe)
 		return 0;
@@ -2258,8 +2364,8 @@ static int bch2_check_stripe_to_lru_ref(struct btree_trans *trans,
 
 int bch2_check_stripe_to_lru_refs(struct btree_trans *trans)
 {
-	struct bkey_buf last_flushed __cleanup(bch2_bkey_buf_exit);
-	bch2_bkey_buf_init(&last_flushed);
+	struct wb_maybe_flush last_flushed __cleanup(wb_maybe_flush_exit);
+	wb_maybe_flush_init(&last_flushed);
 
 	return for_each_btree_key_commit(trans, iter, BTREE_ID_stripes,
 				POS_MIN, BTREE_ITER_prefetch, k,
